# for message type classification see
# http://owfs.org/index.php?page=owserver-message-types
# and 'enum msg_classification' from module/owlib/src/include/ow_message.h

import attr
import struct
from .util import ValueEvent
from .error import _errors, GenericOWFSReplyError

import logging
logger = logging.getLogger(__name__)


class OWMsg:
    """Constants for the owserver api message types."""
    error = 0
    nop = 1
    read = 2
    write = 3
    dir = 4
    size = 5
    presence = 6
    dirall = 7
    get = 8
    dirallslash = 9
    getslash = 10


class OWFlag:
    cache = 0x1  # ?
    busret = 0x2  # enumeration includes bus names
    persist = 0x4  # use persistent connections
    alias = 0x8
    safemode = 0x10
    uncached = 0x20
    ownet = 0x100
    _offset = 0


class OWdevformat:
    fdi = 0
    fi = 1
    fdidc = 2
    fdic = 3
    fidc = 4
    fic = 5
    _offset = 24


class OWtempformat:
    celsius = 0
    fahrenheit = 1
    kelvin = 2
    rankine = 3
    _offset = 16


class OWpressureformat:
    mbar = 0
    atm = 1
    mmHg = 2
    inHg = 3
    psi = 4
    pa = 5
    _offset = 18


class ServerBusy(Exception):
    """Receiver error when the server signals it's busy"""
    pass


class Retry(Exception):
    """Sender resubmitted."""
    pass


master_id = 1
class MessageProtocol:
    MAX_LENGTH = 9999

    def __init__(self, master, is_server=False):
        self.master = master

        global master_id
        self.master_id = master_id
        master_id += 1

        self.stream = master.stream
        self.is_server = is_server
        self._buf = b''
        logger.debug("START %s %s",self.master_id,self.master)

    async def _read_buf(self, nbytes):
        while len(self._buf) < nbytes:
            more = await self.stream.receive_some(4096)
            if not more:
                raise StopAsyncIteration
            self._buf += more
        res = self._buf[:nbytes]
        self._buf = self._buf[nbytes:]
        return res

    def __aiter__(self):
        return self

    async def __anext__(self):
        hdr = await self._read_buf(24)
        version, payload_len, ret_value, format_flags, data_len, offset = struct.unpack('!6i', hdr)
        if offset & 0x8000:
            offset = 0
        if version != 0:
            raise RuntimeError("Wrong version: %d" % (version,))
        if payload_len == -1 and data_len == 0 and offset == 0:
            raise ServerBusy
        if payload_len > self.MAX_LENGTH:
            raise RuntimeError("Server tried to send too much: %d" % (payload_len,))
        if payload_len == 0:
            data_len = 0
        data = await self._read_buf(payload_len)
        logger.debug(
            "OW%s recv%s %x %x %x %x %x %x %s", self.master_id, "S"
            if self.is_server else "", version, payload_len,
            ret_value, format_flags, data_len, offset, repr(data)
        )
        if self.is_server:
            return ret_value, format_flags, data, data_len
        else:
            data = data[:data_len]
            return ret_value, data

    async def write(self, typ, flags, rlen=0, data=b'', offset=0):
        if data is None:
            logger.debug(
                "OW%s send%s %x %x %x %x %x %x -", self.master_id, "S"
                if self.is_server else "", 0, -1, typ, flags, rlen, offset
            )
            await self.stream.send_all(struct.pack("!6i", 0, -1, typ, flags, rlen, offset))
        else:
            logger.debug(
                "OW%s send%s %x %x %x %x %x %x %s", self.master_id, "S"
                if self.is_server else "", 0, len(data), typ, flags, rlen, offset, repr(data)
            )
            await self.stream.send_all(
                struct.pack("!6i", 0, len(data), typ, flags, rlen, offset) + data
            )


class Message:
    timeout = 0.5
    cancelled = False

    def __init__(self, typ, data, rlen):
        #self.persist = persist
        self.typ = typ
        self.data = data
        self.rlen = rlen
        self.event = ValueEvent()

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, repr(self.data))

    async def cancel(self):
        self.cancelled = True
        await self.event.cancel()

    async def write(self, protocol):
        """Send an OWFS message to the other end of the connection.
        """
        # Assume a modern server
        flags = 0
        flags |= OWFlag.persist
        flags |= OWFlag.busret
        flags |= OWFlag.uncached
        flags |= OWFlag.ownet
        flags |= OWtempformat.celsius << OWtempformat._offset
        flags |= OWdevformat.fdidc << OWdevformat._offset
        flags |= OWpressureformat.mbar << OWpressureformat._offset

        # TODO
        if self.event is not None and not self.event.is_set():
            await self.event.set_error(Retry())
        self.event = ValueEvent()

        await protocol.write(self.typ, flags, self.rlen, self.data)

    def _resubmit(self):
        self.event = ValueEvent()

    async def process_reply(self, res, data, server):
        logger.debug("PROCESS %s %s %s", self, res, data)
        if res < 0:
            err = _errors.get(-res, GenericOWFSReplyError)
            if err is GenericOWFSReplyError:
                error = err(res, self, server)
            else:
                error = err(self, server)
            await self.event.set_error(error)
            return
        elif res > 0:
            assert len(data) == res, (data, res)
        data = self._process(data)
        if data is not None:
            await self.event.set(data)

    async def process_error(self, exc):
        await self.event.set_error(exc)

    def _process(self, data):
        return data

    @property
    def get_reply(self):
        return self.event.get

    @property
    def done(self):
        return self.event.is_set

    def _check(self, cmd, data=None):
        """Used in testing: verify that a message is correct."""
        assert self.typ == cmd, (self.typ, cmd)
        if data is not None:
            assert self.data == data, (cmd, self.data, data)


def _path(path):
    """Helper to build an OWFS path from a list"""
    path = '/' + '/'.join(str(x) for x in path)
    return path.encode('utf-8') + b'\0'


class NOPMsg(Message):
    def __init__(self):
        super().__init__(OWMsg.nop, b'', 0)

    def __repr__(self):
        return "<%s>" % (self.__class__.__name__)


class AttrGetMsg(Message):
    """read an OWFS value"""
    timeout = 2

    def __init__(self, *path):
        assert path
        self.path = path
        super().__init__(OWMsg.read, _path(self.path), 8192)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, '/' + '/'.join(self.path))


class AttrSetMsg(Message):
    """write an OWFS value"""
    timeout = 1

    def __init__(self, *path, value):
        assert path is not None
        self.path = path
        self.value = value
        if isinstance(value, bool):
            value = b'1' if value else b'0'
        elif not isinstance(value, bytes):
            value = str(value).encode("utf-8")
        super().__init__(OWMsg.write, _path(self.path) + value, len(value))

    def __repr__(self):
        return "<%s %s =%s>" % (self.__class__.__name__, '/' + '/'.join(self.path), self.value)


class DirMsg(Message):
    """Read an owfs directory"""
    timeout = 10

    def __init__(self, path):
        self.path = path
        p = _path(self.path)
        super().__init__(OWMsg.dirall, p, len(p) - 1)

    def _process(self, data):
        if data == b'':
            return []
        res = []
        for entry in data.split(b","):
            assert b'\0' not in entry
            entry = entry.decode("utf-8")
            s = entry.rfind('/')
            if s > -1:
                entry = entry[s + 1:]
            res.append(entry)
        return res



import trio
import pytest
from copy import deepcopy

from asyncowfs.protocol import NOPMsg, DirMsg, AttrGetMsg, AttrSetMsg
from asyncowfs.error import NoEntryError
from asyncowfs.event import ServerRegistered, ServerConnected, ServerDisconnected, \
    ServerDeregistered, DeviceAdded, DeviceLocated, DeviceNotFound, BusAdded_Path, \
    BusAdded, BusDeleted, DeviceValue, DeviceAlarm
from asyncowfs.bus import Bus

from .mock_server import server, EventChecker
from .structs import structs

import logging
logger = logging.getLogger(__name__)

# We can just use 'async def test_*' to define async tests.
# This also uses a virtual clock fixture, so time passes quickly and
# predictably.

basic_tree = {
    "bus.0":
        {
            "alarm": {},
            "simultaneous": {
                "temperature": 0,
                },
            "10.345678.90":
                {
                    "latesttemp": "12.5",
                    "templow": "15",
                    "temphigh": "20",
                }
        },
    "structure": structs,
}


async def test_alarm(mock_clock):
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            BusAdded,
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            DeviceAlarm("10.345678.90"),
            DeviceValue("10.345678.90", "temperature",12.5),
            #DeviceValue("10.345678.90", "temperature",12.5),
            #DeviceNotFound("10.345678.90"),
            ServerDisconnected,
            DeviceNotFound("10.345678.90"),
            BusDeleted,
            ServerDeregistered,
        ]
    )
    mock_clock.autojump_threshold = 0.1
    my_tree = deepcopy(basic_tree)
    dt = my_tree['bus.0'].pop('10.345678.90')
    async with server(tree=my_tree) as ow: # , polling=True, events=e1) as ow:
        dev = await ow.get_device("10.345678.90")
        dev.interval_alarm = 10
        dev.interval_temperature = 15

        my_tree['bus.0']['10.345678.90'] = dt
        await ow.scan_now()
        my_tree['bus.0']['alarm']['10.345678.90'] = dt
        await trio.sleep(12) # allow alarm poll to trigger
        assert int(dt['temphigh']) == 20
        assert int(dt['templow']) == 11
        assert dev.alarm_temperature == 12.5
        await trio.sleep(5) # allow temperature poll to trigger




"""
Events: whatever is happening on the bus
"""

import attr


class Event:
    """Base class for all events"""
    pass


@attr.s
class ServerEvent(Event):
    """Base class for all server-related events"""
    server = attr.ib()


@attr.s
class ServerRegistered(ServerEvent):
    """A new known server appears. The server is not yet connected!"""
    pass


@attr.s
class ServerConnected(ServerEvent):
    """We have connected to a server"""
    pass


@attr.s
class ServerDisconnected(ServerEvent):
    """We have disconnected from a server"""
    pass


@attr.s
class ServerDeregistered(ServerEvent):
    """This server is no longer known"""
    pass


@attr.s
class BusEvent(Event):
    """Base class for all Bus-related events"""
    bus = attr.ib()


@attr.s
class BusAdded(BusEvent):
    """The Bus has been created. Its location is not yet known!"""
    pass


class BusAdded_Path:
    """Not an event. Used for storing the bus path for comparisons in tests."""

    def __init__(self, *path):
        self.path = path

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, '/'.join(self.path))

    def __eq__(self, x):
        if isinstance(x, BusAdded_Path):
            x = x.path
        elif isinstance(x, BusAdded):
            x = x.bus.path
        elif not isinstance(x, (list, tuple)):
            return False
        return self.path == x


@attr.s
class BusDeleted(BusEvent):
    """The Bus has been deleted"""
    pass


@attr.s
class DeviceEvent(Event):
    """Base class for all device-related events"""
    device = attr.ib()


@attr.s
class DeviceAdded(DeviceEvent):
    """The device has been created. Its location is not yet known!"""
    pass


@attr.s
class DeviceDeleted(DeviceEvent):
    """The device has been deleted"""
    pass


@attr.s
class DeviceLocated(DeviceEvent):
    """The device has been found"""
    pass

    @property
    def server(self):
        return self.device.server

    @property
    def path(self):
        return self.device.path


@attr.s
class DeviceNotFound(DeviceEvent):
    """The device's location is no longer known"""
    pass


@attr.s
class DeviceAlarm(DeviceEvent):
    """The device triggered an alarm condition."""
    reasons = attr.ib(factory=dict)
    pass


@attr.s
class DeviceValue(DeviceEvent):
    """The device poll has read a value."""
    attribute = attr.ib()
    value = attr.ib()




import trio
import pytest
from copy import deepcopy

from asyncowfs.protocol import NOPMsg, DirMsg, AttrGetMsg, AttrSetMsg
from asyncowfs.error import NoEntryError
from asyncowfs.event import ServerRegistered, ServerConnected, ServerDisconnected, \
    ServerDeregistered, DeviceAdded, DeviceLocated, DeviceNotFound, BusAdded_Path, \
    BusAdded, BusDeleted
from asyncowfs.bus import Bus

from .mock_server import server, EventChecker
from .structs import structs

import logging
logger = logging.getLogger(__name__)

# We can just use 'async def test_*' to define async tests.
# This also uses a virtual clock fixture, so time passes quickly and
# predictably.

basic_tree = {
    "bus.0":
        {
            "10.345678.90":
                {
                    "whatever": "hello",
                    "latesttemp": "12.5",
                    "temperature": "12.5",
                    "templow": "10",
                    "temphigh": "20",
                    "foo": {
                        "bar": "42",
                        "baz": {
                            "qux": "99.875",
                        },
                        "plugh.A": 1,
                        "plugh.B": 2,
                        "plugh.C": 3,
                        "plover.0": 7,
                        "plover.1": 8,
                        "plover.2": 9,
                    }
                }
        },
    "structure": structs,
}

coupler_tree = {
    "bus.0":
        {
            "10.345678.90": {
                "whatever": "hello",
                "temperature": "12.5",
            },
            "1F.ABCDEF.F1":
                {
                    "main": {
                        "20.222222.22": {
                            "some": "chip",
                        },
                    },
                    "aux": {
                        "28.282828.28": {
                            "another": "chip",
                        },
                    },
                }
        },
    "structure": structs,
}


async def test_empty_server():
    msgs = [
        NOPMsg(),
        DirMsg(()),
    ]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            ServerDisconnected,
            ServerDeregistered,
        ]
    )
    async with server(msgs=msgs, events=e1):  # as ow:
        await trio.sleep(0)


async def test_send_idle(mock_clock):
    mock_clock.autojump_threshold = 0.1
    msgs = [
        NOPMsg(),
        DirMsg(()),
        NOPMsg(),
    ]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            ServerDisconnected,
            ServerDeregistered,
        ]
    )
    async with server(msgs=msgs, events=e1):  # as ow:
        await trio.sleep(15)


async def test_missing_event():
    msgs = [
        NOPMsg(),
        DirMsg(()),
    ]
    e1 = EventChecker([
        ServerRegistered,
        ServerConnected,
        ServerDisconnected,
    ])
    with pytest.raises(RuntimeError) as r:
        async with server(msgs=msgs, events=e1):  # as ow:
            await trio.sleep(0)
    assert "Unexpected event " in r.value.args[0]


async def test_more_event():
    msgs = [
        NOPMsg(),
        DirMsg(()),
    ]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            ServerDisconnected,
            ServerDeregistered,
            ServerDeregistered,
        ]
    )
    with pytest.raises(RuntimeError) as r:
        async with server(msgs=msgs, events=e1):  # as ow:
            await trio.sleep(0)
    assert "Superfluous event " in r.value.args[0]


async def test_bad_event():
    msgs = [
        NOPMsg(),
        DirMsg(()),
    ]
    e1 = EventChecker([
        ServerRegistered,
        ServerConnected,
        ServerConnected,
        ServerDeregistered,
    ])
    with pytest.raises(RuntimeError) as r:
        async with server(msgs=msgs, events=e1):  # as ow:
            await trio.sleep(0)
    assert "Wrong event: want " in r.value.args[0]


async def test_wrong_msg():
    msgs = [
        NOPMsg(),
        DirMsg(("foobar",)),
    ]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            ServerDisconnected,
            ServerDeregistered,
        ]
    )
    with pytest.raises(RuntimeError) as r:
        async with server(msgs=msgs, events=e1):  # as ow:
            await trio.sleep(0)
    assert " wrong seen " in r.value.args[0]


async def test_wrong_msg_2():
    msgs = [
        NOPMsg(),
        NOPMsg(),
    ]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            ServerDisconnected,
            ServerConnected,
            ServerDisconnected,
            ServerDeregistered,
        ]
    )
    with pytest.raises(RuntimeError) as r:
        async with server(msgs=msgs, events=e1):  # as ow:
            await trio.sleep(0)
    assert " wrong seen " in r.value.args[0]


async def test_missing_msg():
    msgs = [
        NOPMsg(),
    ]
    e1 = EventChecker([
        ServerRegistered,
        ServerConnected,
        ServerDisconnected,
    ])
    with pytest.raises(RuntimeError) as r:
        async with server(msgs=msgs, events=e1):  # as ow:
            await trio.sleep(0)
    assert "Unexpected command " in r.value.args[0]


async def test_more_msg():
    msgs = [
        NOPMsg(),
        DirMsg(()),
        DirMsg(()),
    ]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            ServerDisconnected,
            ServerDeregistered,
        ]
    )
    with pytest.raises(RuntimeError) as r:
        async with server(msgs=msgs, events=e1):  # as ow:
            await trio.sleep(0)
    assert " not seen " in r.value.args[0]


async def test_basic_server():
    msgs = [
        NOPMsg(),
        DirMsg(()),
        DirMsg(("bus.0",)),
        AttrGetMsg("bus.0", "10.345678.90", "temperature"),
        AttrSetMsg("bus.0", "10.345678.90", "temperature", value=98.25),
        AttrGetMsg("bus.0", "10.345678.90", "temperature"),
        AttrSetMsg("bus.0", "10.345678.90", "temperature", value=12.5),
    ]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            BusAdded_Path("bus.0"),
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            ServerDisconnected,
            DeviceNotFound,
            BusDeleted,
            ServerDeregistered,
        ]
    )
    async with server(tree=basic_tree) as ow: # msgs=msgs, events=e1) as ow:
        await trio.sleep(0)
        dev = await ow.get_device("10.345678.90")
        assert dev.bus == ('bus.0',)
        assert isinstance(dev.bus, Bus)
        assert float(await dev.attr_get("temperature")) == 12.5
        await dev.attr_set("temperature", value=98.25)
        assert float(await dev.attr_get("temperature")) == 98.25
        await dev.attr_set("temperature", value=12.5)


async def test_basic_structs(mock_clock):
    mock_clock.autojump_threshold = 0.1
    async with server(tree=basic_tree, options={'slow_every': [0, 1], 'busy_every': [0, 0, 1],
                                                'close_every': [0, 0, 0, 1]}) as ow:
        await trio.sleep(0)
        dev = await ow.get_device("10.345678.90")
        await ow.ensure_struct(dev)
        await dev.wait_bus()
        assert await dev.temperature == 12.5
        await dev.set_temphigh(98.25)
        assert await dev.temphigh == 98.25

        # while we're at it, test our ability to do things in parallel on a broken server
        dat = {}
        evt = trio.Event()

        async def get_val(tag):
            await evt.wait()
            dat[tag] = await getattr(dev, tag)

        async with trio.open_nursery() as n:
            n.start_soon(get_val, 'temperature')
            n.start_soon(get_val, 'temphigh')
            n.start_soon(get_val, 'templow')
            await trio.sleep(1)
            evt.set()
        assert dat == {'temphigh': 98.25, 'temperature': 12.5, 'templow': 10.0}, dat


async def test_more_structs(mock_clock):
    mock_clock.autojump_threshold = 0.1
    async with server(tree=basic_tree) as ow:
        await trio.sleep(0)
        dev = await ow.get_device("10.345678.90")
        await ow.ensure_struct(dev)
        assert await dev.temperature == 12.5
        await dev.set_temphigh(98.25)
        assert await dev.temphigh == 98.25
        assert await dev.foo.get_bar() == 42
        assert await dev.foo.baz.qux == 99.875
        assert await dev.foo.plugh[0] == 1
        assert await dev.foo.plugh[1] == 2
        assert await dev.foo.plugh[2] == 3
        assert await dev.foo.plover[0] == 7
        assert await dev.foo.get_plover(1) == 8
        assert await dev.foo.plover[2] == 9
        await dev.foo.set_bar(123)
        await dev.foo.baz.set_qux(234)
        await dev.foo.set_plugh(1, 11)
        await dev.foo.set_plover(1, 111)
        assert await dev.foo.bar == 123
        assert await dev.foo.baz.qux == 234
        assert await dev.foo.plugh[0] == 1
        assert await dev.foo.plugh[1] == 11
        assert await dev.foo.plugh[2] == 3
        assert await dev.foo.plover[0] == 7
        assert await dev.foo.plover[1] == 111
        assert await dev.foo.get_plover(2) == 9

async def test_coupler_server():
    msgs = [
        NOPMsg(),
        DirMsg(()),
        DirMsg(("bus.0",)),
        DirMsg(("bus.0", "1F.ABCDEF.F1", "main")),
        DirMsg(("bus.0", "1F.ABCDEF.F1", "aux")),
    ]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            BusAdded_Path("bus.0"),
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            DeviceAdded("1F.ABCDEF.F1"),
            DeviceLocated("1F.ABCDEF.F1"),
            BusAdded_Path("bus.0", "1F.ABCDEF.F1", "main"),
            DeviceAdded("20.222222.22"),
            DeviceLocated("20.222222.22"),
            BusAdded_Path("bus.0", "1F.ABCDEF.F1", "aux"),
            DeviceAdded("28.282828.28"),
            DeviceLocated("28.282828.28"),
            ServerDisconnected,
            DeviceNotFound,
            BusDeleted,
            DeviceNotFound,
            BusDeleted,
            DeviceNotFound,
            DeviceNotFound,
            BusDeleted,
            ServerDeregistered,
        ]
    )
    async with server(tree=coupler_tree): # msgs=msgs, events=e1, tree=coupler_tree):  # as ow:
        await trio.sleep(0)


async def test_wrong_bus():
    msgs = [
        NOPMsg(),
        DirMsg(()),
        DirMsg(("bus.0",)),
    ]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            BusAdded_Path("bus-whatever"),
            #       DeviceAdded("10.345678.90"),
            #       DeviceLocated("10.345678.90"),
            #       ServerDisconnected,
            #       ServerDeregistered,
        ]
    )
    with pytest.raises(RuntimeError) as r:
        async with server(msgs=msgs, events=e1, tree=basic_tree):  # as ow:
            await trio.sleep(0)
    assert "Wrong event: want " in r.value.args[0]


async def test_slow_server(mock_clock):
    mock_clock.autojump_threshold = 0.1
    msgs = [
        [
            NOPMsg(),
            DirMsg(()),
            NOPMsg(),
        ], [
            DirMsg(()),
            NOPMsg(),
            DirMsg(("bus.0",)),
            NOPMsg(),
        ], [
            DirMsg(("bus.0",)),
            NOPMsg(),
        ], [
            DirMsg(("bus.0",)),
        ]
    ]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            ServerDisconnected,
            ServerConnected,
            BusAdded,
            ServerDisconnected,
            ServerConnected,
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            ServerDisconnected,
            DeviceNotFound("10.345678.90"),
            BusDeleted,
            ServerDeregistered,
        ]
    )
    async with server(msgs=msgs, events=e1, tree=basic_tree,
                      options={'slow_every': [0, 0, 15, 0, 0]}):  # as ow:
        await trio.sleep(0)


async def test_busy_server():
    msgs = [
        NOPMsg(),
        DirMsg(()),
        DirMsg(("bus.0",)),
    ]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            BusAdded,
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            ServerDisconnected,
            DeviceNotFound("10.345678.90"),
            BusDeleted,
            ServerDeregistered,
        ]
    )
    async with server(msgs=msgs, events=e1, tree=basic_tree, options={'busy_every': [0, 0, 1]}):
        await trio.sleep(0)
    async with server(msgs=msgs, events=e1, tree=basic_tree, options={'busy_every': [1, 0, 1]}):
        await trio.sleep(0)


async def test_disconnecting_server():
    msgs = [[
        NOPMsg(),
        DirMsg(()),
    ], [
        DirMsg(()),
        DirMsg(("bus.0",)),
    ], [
        DirMsg(("bus.0",)),
    ]]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            ServerDisconnected,
            ServerConnected,
            BusAdded,
            ServerDisconnected,
            ServerConnected,
            DeviceAdded("10.345678.90"),
            ServerDisconnected,
            ServerConnected,
            DeviceLocated("10.345678.90"),
            ServerDisconnected,
            DeviceNotFound("10.345678.90"),
            BusDeleted,
            ServerDeregistered,
        ]
    )
    async with server(tree=basic_tree, # msgs=msgs, # events=e1,
                      options={'close_every': [0, 0, 0, 1]}):  # as ow:
        await trio.sleep(0)


async def test_disconnecting_server_2(mock_clock):
    mock_clock.autojump_threshold = 0.1
    msgs = [[], [
        NOPMsg(),
        DirMsg(()),
    ], [
        DirMsg(()),
        DirMsg(("bus.0",)),
    ], [
        DirMsg(("bus.0",)),
    ]]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            ServerDisconnected,
            ServerConnected,
            ServerDisconnected,
            ServerConnected,
            BusAdded,
            ServerDisconnected,
            ServerConnected,
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            ServerDisconnected,
            DeviceNotFound("10.345678.90"),
            BusDeleted,
            ServerDeregistered,
        ]
    )
    async with server(tree=basic_tree, # msgs=msgs, events=e1, tree=basic_tree,
                      options={'close_every': [0, 0, 1, 0, 0]}):  # as ow:
        await trio.sleep(0)


async def test_dropped_device(mock_clock):
    mock_clock.autojump_threshold = 0.1
    msgs = [
        NOPMsg(),
        DirMsg(()),
        DirMsg(("bus.0",)),
        DirMsg(()),
        DirMsg(("bus.0",)),
        DirMsg(()),
        DirMsg(("bus.0",)),
        DirMsg(()),
        DirMsg(("bus.0",)),
        DirMsg(()),
        DirMsg(("bus.0",)),
    ]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            BusAdded,
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            DeviceNotFound("10.345678.90"),
            ServerDisconnected,
            BusDeleted,
            ServerDeregistered,
        ]
    )
    my_tree = deepcopy(basic_tree)
    async with server(tree=my_tree, msgs=msgs, events=e1) as ow:
        dev = await ow.get_device('10.345678.90')
        assert dev.bus is not None
        del my_tree['bus.0']['10.345678.90']
        assert dev._unseen == 0

        await ow.scan_now(polling=False)
        assert dev._unseen == 1
        await ow.scan_now(polling=False)
        assert dev._unseen == 2
        await ow.scan_now(polling=False)
        assert dev._unseen == 3
        assert dev.bus is not None
        await ow.scan_now(polling=False)
        assert dev._unseen == 3
        assert dev.bus is None


async def test_manual_device(mock_clock):
    class Checkpoint:
        pass

    mock_clock.autojump_threshold = 0.1
    msgs = [
        NOPMsg(),
        DirMsg(()),
        DirMsg(("bus.0",)),
        NOPMsg(),
        DirMsg(()),
        DirMsg(("bus.0",)),
    ]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            BusAdded,
            Checkpoint,
            DeviceAdded("10.345678.90"),
            Checkpoint,
            DeviceLocated("10.345678.90"),
            ServerDisconnected,
            DeviceNotFound("10.345678.90"),
            BusDeleted,
            ServerDeregistered,
        ]
    )
    my_tree = deepcopy(basic_tree)
    entry = my_tree['bus.0'].pop('10.345678.90')
    async with server(msgs=msgs, events=e1, tree=my_tree) as ow:
        await ow.push_event(Checkpoint())
        dev = await ow.get_device('10.345678.90')
        assert dev.bus is None
        await trio.sleep(15)
        await ow.push_event(Checkpoint())
        my_tree['bus.0']['10.345678.90'] = entry
        await ow.scan_now(polling=False)
        assert dev.bus is not None


async def test_nonexistent_device(mock_clock):
    mock_clock.autojump_threshold = 0.1
    my_tree = deepcopy(basic_tree)
    entry = my_tree.pop('bus.0')
    async with server(tree=my_tree, scan=None) as ow:
        bus = await ow.test_server.get_bus('bus.0')
        assert bus.server == ow.test_server

        dev = await ow.get_device('10.345678.90')
        await dev.locate(bus)
        with pytest.raises(NoEntryError):
            assert float(await dev.attr_get("temperature")) == 12.5


async def test_manual_bus(mock_clock):
    class Checkpoint:
        pass

    mock_clock.autojump_threshold = 0.1
    msgs = [
        NOPMsg(),
        NOPMsg(),
        AttrGetMsg("bus.0", "10.345678.90", "temperature"),
        NOPMsg(),
        DirMsg(()),
        DirMsg(("bus.0",)),
    ]
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            BusAdded_Path("bus.0"),
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            Checkpoint,
            Checkpoint,
            ServerDisconnected,
            DeviceNotFound("10.345678.90"),
            BusDeleted,
            ServerDeregistered,
        ]
    )
    async with server(  #msgs=msgs, events=e1,
            tree=basic_tree, scan=None, initial_scan=False) as ow:
        bus = await ow.test_server.get_bus('bus.0')
        assert bus.server == ow.test_server

        dev = await ow.get_device('10.345678.90')
        assert dev.bus is None
        await dev.locate(bus)
        assert dev.bus is not None
        assert float(await dev.attr_get("temperature")) == 12.5



# base implementation

import anyio
from functools import partial
try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager

from typing import Optional,Union

from .server import Server
from .device import Device
from .event import ServerRegistered, ServerDeregistered
from .event import DeviceAdded, DeviceDeleted
from .util import ValueEvent

import logging
logger = logging.getLogger(__name__)

__all__ = ["OWFS"]


class Service:
    """\
        This is the master class you use for communicating with OWFS.
        You typically start it thus::

            async def rdr(ow):
                async for evt in ow:
                    process(evt)

            async with OWFS() as ow:
                await ow.add_task(rdr, ow)
                s = await ow.add_server("localhost",4304")

        :param scan: time between directory scanning.
            None: do not scan repeatedly

        :param initial_scan: time to first scan
            False: no initial scal
            True: scan immediately, block before returning

        :param polling: flag whether to poll devices.
            Default: True.

        :param load_structs: Flag whether to generate accessors from OWFS data.
            Default: True
        """

    def __init__(self, nursery, scan: Union[float,None] = None, initial_scan: Union[float,bool] = True, load_structs: bool = True, polling: bool = True, random: int = 0):
        self.nursery = nursery
        self._servers = set()  # typ.MutableSet[Server]  # Server
        self._devices = dict()  # ID => Device
        self._tasks = set()  # typ.MutableSet[]  # actually their cancel scopes
        self._event_queue = None  # typ.Optional[anyio.Queue]
        self._random = random
        self._scan = scan
        self._initial_scan = initial_scan
        self._polling = polling
        self._load_structs = load_structs

    async def add_server(self, host: str, port: int = 4304, polling: Optional[bool] = None,
            scan: Union[float,bool,None] = None, initial_scan: Union[float,bool,None] = None,
            random: Optional[int] = None, background: bool = False):
        """Add this server to the list.
        
        :param polling: if False, don't poll.
        :param scan: Override ``self._scan`` for this server.
        :param initial_scan: Override ``self._initial_scan`` for this server.
        """
        if scan is None:
            scan = self._scan
        if initial_scan is None:
            initial_scan = self._initial_scan
        if polling is None:
            polling = self._polling
        if random is None:
            random = self._random

        s = Server(self, host, port)
        await self.push_event(ServerRegistered(s))
        try:
            await s.start(background=background)
        except BaseException as exc:
            logger.error("Could not start %s:%s", host,port)
            await self.push_event(ServerDeregistered(s))
            raise
        self._servers.add(s)
        await s.start_scan(scan=scan, initial_scan=initial_scan, polling=polling, random=random)
        return s

    async def ensure_struct(self, dev, server=None, maybe=False):
        """
        Load a device's class's structure definition from any server.

        :param dev: The device whose class to set up
        :param server: Try this server, not all of them
        :param maybe: if set, don't load if ``load_structs`` is set

        """
        if maybe and not self._load_structs:
            return
        cls = type(dev)
        if cls._did_setup:
            return
        if server is not None:
            await cls.setup_struct(server)
        else:
            for s in list(self._servers):
                await cls.setup_struct(s)
                return

    async def get_device(self, id):
        """
        Return the :class:`anyio_owfs.device.Device` instance for the device
        with this ID. Create it if it doesn't exist (this will trigger a
        `DeviceAdded` event).
        """
        try:
            return self._devices[id]
        except KeyError:
            dev = Device(self, id)
            self._devices[dev.id] = dev
            await self.push_event(DeviceAdded(dev))
            return dev

    async def _add_task(self, val, proc, *args):
        async with anyio.open_cancel_scope() as scope:
            await val.set(scope)
            try:
                await proc(*args)
            finally:
                try:
                    self._tasks.remove(scope)
                except KeyError:
                    pass

    async def scan_now(self, polling=True):
        """
        Task to scan the whole system.
        
        :param polling: if False, do not add polling tasks
        """
        async with anyio.create_task_group() as n:
            for s in list(self._servers):
                await n.spawn(partial(s.scan_now, polling=polling))

    async def add_task(self, proc, *args):
        """
        Add a background task. It is auto-cancelled when the service ends.
        Alternately, this call returns its cancel scope.
        """
        val = ValueEvent()
        await self.nursery.spawn(self._add_task, val, proc, *args)
        scope = await val.get()
        self._tasks.add(scope)
        return scope

    async def push_event(self, event):
        if self._event_queue is not None:
            await self._event_queue.put(event)

    async def _del_server(self, s):
        self._servers.remove(s)
        await self.push_event(ServerDeregistered(s))

    async def _del_device(self, d):
        self._devices.remove(d)
        await self.push_event(DeviceDeleted(d))

    @property
    def devices(self):
        return self._devices.values()

    # context

    async def __aenter__(self):
        return self

    async def __aexit__(self, *tb):
        for s in list(self._servers):
            await s.drop()
        if self._event_queue is not None:
            await self._event_queue.put(None)
        for t in list(self._tasks):
            await t.cancel()

    # listen to events

    @property
    def events(self):
        class EventWrapper:
            async def __aenter__(slf):
                assert self._event_queue is None
                self._event_queue = anyio.create_queue(1000)  # bus events
                return slf

            async def __aexit__(slf, *tb):
                if tb[1] is None and not self._event_queue.empty():
                    async with anyio.open_cancel_scope(shield=True):
                        while not self._event_queue.empty():
                            evt = await self._event_queue.get()
                            if evt is not None:
                                logger.error("Unprocessed: %s",evt)
                self._event_queue = None

            def __aiter__(slf):
                return slf

            async def __anext__(slf):
                if self._event_queue is None:
                    raise StopAsyncIteration
                res = await self._event_queue.get()
                if res is None:
                    raise StopAsyncIteration
                return res

        return EventWrapper()


@asynccontextmanager
async def OWFS(**kwargs):
    async with anyio.create_task_group() as n:
        s = Service(n, **kwargs)
        async with s:
            yield s



"""
Buses.
"""

from random import random
import anyio

from .device import NotADevice, split_id, NoLocationKnown
from .event import BusAdded, BusDeleted, DeviceAlarm
from .error import OWFSReplyError

import logging
logger = logging.getLogger(__name__)


class Bus:
    """Describes one bus.
    """

    def __init__(self, server, *path):
        self.service = server.service
        self.server = server
        self.path = path

        self._buses = dict()  # subpath => bus
        self._devices = dict()  # id => device
        self._unseen = 0  # didn't find when scanning
        self._tasks = dict()  # polltype => task
        self._intervals = dict()
        self._random = dict() # varying intervals

    def __repr__(self):
        return "<%s:%s %s>" % (self.__class__.__name__, self.server, '/' + '/'.join(self.path))

    def __eq__(self, x):
        x = getattr(x, 'path', x)
        return self.path == x

    def __hash__(self):
        return hash(self.path)

    @property
    def devices(self):
        """Iterate over the devices on this bus"""
        try:
            return list(self._devices.values())
        except AttributeError:
            return ()

    @property
    def buses(self):
        """Iterate over the sub.buses on this bus"""
        return list(self._buses.values())

    async def delocate(self):
        """This bus can no longer be found"""
        if self._buses:
            for b in self.buses:
                await b.delocate()
            self._buses = None
        if self._devices:
            for d in self.devices:
                await d.delocate(bus=self)
            self._devices = None
        await self.service.push_event(BusDeleted(self))

    @property
    def all_buses(self):
        yield self
        for b in self.buses:
            yield from b.all_buses

    async def get_bus(self, *path):
        try:
            return self._buses[path]
        except TypeError:
            return None # bus is gone
        except KeyError:
            bus = Bus(self.server, *self.path, *path)
            self._buses[path] = bus
            await self.service.push_event(BusAdded(bus))
            return bus

    async def _scan_one(self, polling=True):
        """Scan a single bus, plus all buses attached to it"""
        buses = set()
        res = await self.dir()
        old_devs = set(self._devices.keys())
        for d in res:
            try:
                split_id(d)  # only tests for conformity
            except NotADevice as err:
                logger.debug("Not a device: %s", err)
                continue
            dev = await self.service.get_device(d)
            await self.service.ensure_struct(dev, server=self.server, maybe=True)
            if dev.bus is self:
                old_devs.remove(d)
            else:
                await self.add_device(dev)
            dev._unseen = 0
            logger.debug("Found %s/%s", '/'.join(self.path), d)
            for b in dev.buses():
                buses.add(b)
                bus = await self.get_bus(*b)
                if bus is not None:
                    buses.update(await bus._scan_one(polling=polling))

        for d in old_devs:
            dev = self._devices[d]
            if dev._unseen > 2:
                await dev.delocate(self)
            else:
                dev._unseen += 1
        if polling:
            await self.update_poll()
        return buses

    async def update_poll(self):
        """Start all new polling jobs, terminate old ones"""
        items = set()
        intervals = dict()
        randoms = dict()
        for dev in self.devices:
            for k in dev.polling_items():
                i = dev.polling_interval(k)
                if i is None:
                    continue
                items.add(k)
                if isinstance(i, (tuple,list)):
                    i,j = i
                else:
                    j = None
                oi = intervals.get(k, i)
                intervals[k] = min(oi, i)
                if j is not None:
                    oi = randoms.get(k, j)
                    randoms[k] = min(oi, j)

        old_items = set(self._tasks.keys()) - items
        for x in old_items:
            j = self._tasks.pop(x)
            await j.cancel()

        self._intervals.update(intervals)
        self._random.update(randoms)
        for x in items:
            if x not in self._tasks:
                self._tasks[x] = await self.service.add_task(self._poll, x)

    async def _poll(self, name):
        """Task to run a specific poll in the background"""
        while True:
            i = self._intervals[name]
            j = self._random.get(name, 0)
            if j:
                i *= (1+(random()-0.5)/j)
            logger.info("Delay %s for %f" % (name,i))
            await anyio.sleep(i)
            await self.poll(name)

    async def add_device(self, dev):
        await dev.locate(self)
        self._devices[dev.id] = dev

    async def _del_device(self, dev):
        del self._devices[dev.id]

    def dir(self, *subpath):
        return self.server.dir(*self.path, *subpath)

    async def attr_get(self, *attr):
        """Read this attribute"""
        return await self.server.attr_get(*self.path, *attr)

    async def attr_set(self, *attr, value):
        """Write this attribute"""
        if self.server is None:
            raise NoLocationKnown(self)
        return await self.server.attr_set(*self.path, *attr, value=value)

    ### Support for polling and alarm handling

    async def poll(self, name):
        """Run one poll.

        This typically runs via a :meth:`_poll` task, started by :meth:`update_poll`.
        """
        try:
            try:
                p = getattr(self, 'poll_'+name)
            except AttributeError:
                for d in self.devices:
                    p = getattr(d, 'poll_'+name, None)
                    if p is not None:
                        await p()
            else:
                await p()
        except OWFSReplyError as exc:
            logger.exception("Poll '%s' on %s", name, self)

    async def poll_alarm(self):
        """Scan the 'alarm' subdirectory"""
        for dev in await self.dir("alarm"):
            dev = await self.service.get_device(dev)
            await self.add_device(dev)
            reasons = await dev.poll_alarm()
            await self.service.push_event(DeviceAlarm(dev, reasons))

    async def _poll_simul(self, name, delay):
        """Write to a single 'simultaneous' entry"""
        await self.attr_set("simultaneous", name, value=1)
        await anyio.sleep(delay)
        for dev in self.devices:
            try:
                p = getattr(dev, "poll_"+name)
            except AttributeError:
                pass
            else:
                await p()

    def poll_temperature(self):
        """Read all temperature data"""
        return self._poll_simul("temperature", 1.2)




"""
Access to an owserver.
"""

import anyio
from anyio.exceptions import IncompleteRead,ClosedResourceError
from collections import deque
from random import random
from typing import Union
from functools import partial
from concurrent.futures import CancelledError

from .event import ServerConnected, ServerDisconnected
from .event import BusAdded
from .protocol import NOPMsg, DirMsg, AttrGetMsg, AttrSetMsg, MessageProtocol, ServerBusy, Retry
from .bus import Bus
from .util import ValueEvent

import logging
logger = logging.getLogger(__name__)


class Server:
    """\
        Encapsulate one server connection.
    """

    def __init__(self, service, host="localhost", port=4304):
        self.service = service
        self.host = host
        self.port = port
        self.stream = None
        self._msg_proto = None
        self.requests = deque()
        self._wqueue = anyio.create_queue(100)
        self._read_task = None
        self._write_task = None
        self._scan_task = None
        self._buses = dict()  # path => bus
        self._scan_lock = anyio.create_lock()
        self._scan_args = {}

    async def get_bus(self, *path):
        """Return the bus at this path. Allocate new if not existing."""
        try:
            return self._buses[path]
        except KeyError:
            bus = Bus(self, *path)
            self._buses[bus.path] = bus
            await self.service.push_event(BusAdded(bus))
            return bus

    def __repr__(self):
        return "<%s:%s:%d %s>" % (
            self.__class__.__name__, self.host, self.port, "OK" if self.stream else "closed"
        )

    async def _reader(self, val):
        async with anyio.open_cancel_scope() as scope:
            await val.set(scope)
            it = self._msg_proto.__aiter__()
            while True:
                try:
                    async with anyio.fail_after(15):
                        res, data = await it.__anext__()
                except ServerBusy as exc:
                    logger.info("Server %s busy", self.host)
                except (StopAsyncIteration, TimeoutError, IncompleteRead, ConnectionResetError, ClosedResourceError):
                    await self._reconnect()
                    it = self._msg_proto.__aiter__()
                else:
                    msg = self.requests.popleft()
                    await msg.process_reply(res, data, self)
                    if not msg.done():
                        self.requests.appendleft(msg)

    async def _reconnect(self):
        await self.service.push_event(ServerDisconnected(self))
        if self._write_task is not None:
            await self._write_task.cancel()
            self._write_task = None
        if self._scan_task is not None:
            await self._scan_task.cancel()
            self._scan_task = None
        await self.stream.close()
        self.stream = None

        backoff = 0.2

        while True:
            try:
                self.stream = await anyio.connect_tcp(self.host, self.port)
            except OSError:
                await anyio.sleep(backoff)
                if backoff < 10:
                    backoff *= 1.5
            except BaseException as exc:
                logger.exception("Owch")
            else:
                self._msg_proto = MessageProtocol(self, is_server=False)
                # re-send messages, but skip those that have been cancelled
                logger.warning("Server %s restarting", self.host)
                ml, self.requests = list(self.requests), deque()
                self._wqueue = anyio.create_queue(100)
                await self.service.push_event(ServerConnected(self))
                v_w = ValueEvent()
                await self.service.nursery.spawn(self._writer, v_w)
                self._write_task = await v_w.get()
                for msg in ml:
                    if not msg.cancelled:
                        await self._wqueue.put(msg)
                await self.service.nursery.spawn(partial(self.start_scan,**self._scan_args))
                return

    async def start(self, background=None):
        """Start talking. Returns when the connection is established,
        raises an error if that's not possible.

        TODO: if the connection subsequently drops, it's re-established
        transparently.
        """
        if background:
            await self.service.add_task(self._start)
            return
            
        if self.stream is not None:
            if background is False:
                return
            raise RuntimeError("already open")
        self.stream = await anyio.connect_tcp(self.host, self.port)
        self._msg_proto = MessageProtocol(self, is_server=False)
        await self.service.push_event(ServerConnected(self))
        v_w = ValueEvent()
        v_r = ValueEvent()
        await self.service.nursery.spawn(self._writer, v_w)
        await self.service.nursery.spawn(self._reader, v_r)
        self._write_task = await v_w.get()
        self._read_task = await v_r.get()

        try:
            await self.chat(NOPMsg(), fail=True)
        except BaseException:
            await self.aclose()
            raise

    async def _start(self):
        # try to start repeatedly
        dly = 0.2
        while True:
            try:
                await self.start(background=False)
            except ConnectionRefusedError as exc:
                logger.warning("Connection to %s:%s failed, will retry",
                        self.host,self.port)
            else:
                await self.start_scan(**self._scan_args)
                return
            dly *= 1.5
            if dly > 10:
                dly = 10
            await anyio.sleep(dly)

    async def setup_struct(self, dev):
        await dev.setup_struct(self)

    async def chat(self, msg, fail=False):
        backoff = 0.1
        await self._wqueue.put(msg)
        try:
            while True:
                try:
                    res = await msg.get_reply()
                    return res
                except ServerBusy:
                    await anyio.sleep(backoff)
                    if backoff < 2:
                        backoff *= 1.5
                    msg._resubmit()
                    await self._wqueue.put(msg)
                except Retry:
                    # The message has been repeated.
                    pass
        except BaseException as exc:
            await msg.cancel()
            raise

    async def _writer(self, val):
        async with anyio.open_cancel_scope() as scope:
            await val.set(scope)
            while True:
                try:
                    async with anyio.fail_after(10):
                        msg = await self._wqueue.get()
                except TimeoutError:
                    msg = NOPMsg()

                self.requests.append(msg)
                try:
                    await msg.write(self._msg_proto)
#                except trio.ClosedResourceError:
#                    # will get restarted by .reconnect()
#                    return
                except EnvironmentError as err:
                    logger.warning("Write error: %r", err)
                    return  # will be noticed by the reader
                except IncompleteRead:
                    await self.stream.close()
                    return  # wil be restarted by the reader
                except BaseException:
                    async with anyio.open_cancel_scope(shield=True):
                        await self.aclose()
                    raise

    async def drop(self):
        """Stop talking and delete yourself"""
        try:
            await self.aclose()
        finally:
            await self.service._del_server(self)

    async def aclose(self):
        if self._scan_task is not None:
            await self._scan_task.cancel()
            self._scan_task = None
        if self._write_task is not None:
            await self._write_task.cancel()
            self._write_task = None
        if self._read_task is not None:
            await self._read_task.cancel()
            self._read_task = None

        if self.stream is None:
            return

        try:
            await self.stream.close()
        finally:
            self.stream = None
            await self.service.push_event(ServerDisconnected(self))

        for b in list(self._buses.values()):
            await b.delocate()
        self._buses = None
        for m in self.requests:
            await m.cancel()

    @property
    def all_buses(self):
        for b in list(self._buses.values()):
            yield from b.all_buses

    async def dir(self, *path):
        return await self.chat(DirMsg(path))

    async def _scan(self, interval, initial_interval, polling, random=0):
        if not initial_interval:
            initial_interval = interval
        # 5% variation, to prevent clustering
        if random:
            initial_interval *= 1+(random()-0.5)/random
        await anyio.sleep(initial_interval)

        while True:
            async with self._scan_lock:
                await self.scan_now(polling=polling)
            if not interval:
                return
            i = interval
            if random:
                i *= 1+(random()-0.5)/random
            await anyio.sleep(i)

    async def scan_now(self, polling=True):
        if self._scan_lock.locked():
            # scan in progress: just wait for it to finish
            async with self._scan_lock:
                pass
        else:
            async with self._scan_lock:
                await self._scan_base(polling=polling)

    async def _scan_base(self, polling=True):
        old_paths = set()

        # step 1: enumerate
        try:
            for d in await self.dir():
                if d.startswith("bus."):
                    bus = await self.get_bus(d)
                    bus._unseen = 0
                    try:
                        old_paths.remove(d)
                    except KeyError:
                        pass
                    buses = await bus._scan_one(polling=polling)
                    old_paths -= buses
        except CancelledError:
            return

        # step 2: deregister buses, if not seen often enough
        for p in old_paths:
            bus = self._buses.get(p, None)
            if bus is None:
                continue
            if bus._unseen > 2:
                await bus.delocate()
            else:
                bus._unseen += 1

    async def start_scan(self, scan: Union[float,None] = None,
            initial_scan: Union[float,bool] = True, polling = True,
            random: int = 0):
        """Scan this server.

        :param scan: Flag how often to re-scan the bus.
            None: don't scan at all
            >0: repeat in the background
        :param initial_scan: Flag when to initially scan the bus.
            False: don't.
            True: immediately, wait until complete.
            >0: return immediately, delay initial scan that many seconds.
        :type scan: :class:`float` or ``None``
        :type initial_scan: :class:`float` or :class:`bool`
        :param polling: Flag whether to start tasks for periodic polling
            (alarm handling, temperature, …). Defaults to ``True``.
        """
        self._scan_args = dict(scan=scan,initial_scan=initial_scan,polling=polling,random=random)
        if self.stream is None:
            return
        if not scan and not initial_scan:
            return
        if scan and scan < 1:
            raise RuntimeError("You can't scan that often.")
        if initial_scan is True:
            await self.scan_now(polling=polling)
            initial_scan = False
        if initial_scan or scan:
            self._scan_task = await self.service.add_task(self._scan, scan, initial_scan, polling, random)

    async def attr_get(self, *path):
        return await self.chat(AttrGetMsg(*path))

    async def attr_set(self, *path, value):
        return await self.chat(AttrSetMsg(*path, value=value))



"""walk.py -- an AsyncOWFS implementation of owget

This programs connects to an owserver, enumerates all its nuses, and lists
the devices on it.

With ``-d device``, list that device's attributes. (You don't need the
complete path -- just the ID is sufficient.)

Usage examples:

python3 walk.py
python3 walk.py -d 05.67C6697351FF.BF

"""

from __future__ import print_function

import sys

import collections

import asyncclick as click
from asyncowfs import OWFS

import logging
logger = logging.getLogger('examples.walk')

__all__ = ['main']

async def mon(ow):
    async with ow.events as events:
        async for msg in events:
            logger.info("%s", msg)

@click.command()
@click.option('--host', '-h', default='localhost', help='host running owserver')
@click.option('--port', '-p', default=4304, type=int, help='owserver port')
@click.option('--debug', '-D', is_flag=True, help='Show debug information')
@click.option('--device', '-d', help='List attributes of this device')
async def main(host, port, debug, device):
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)

    async with OWFS() as ow:
        if debug:
            await ow.add_task(mon, ow)
        s = await ow.add_server(host, port)
        if device:
            device = device.rsplit('/',1)[-1]
            d = await ow.get_device(device)
            for f in d.fields:
                print(f)
        else:
            for b in s.all_buses:
                for d in b.devices:
                    p = '/'+'/'.join(d.bus.path)+'/'+d.id
                    print(p)

if __name__ == '__main__':
    main()



"""
Devices.
"""

import attr
import anyio
from functools import partial
from typing import List, Optional

from .event import DeviceLocated, DeviceNotFound, DeviceValue
from .error import IsDirError

import logging
logger = logging.getLogger(__name__)

__all__ = ["Device"]


@attr.s
class NoLocationKnown(RuntimeError):
    device = attr.ib()


@attr.s
class NotADevice(RuntimeError):
    id = attr.ib()


dev_classes = dict()


def register(cls):
    dev_classes[cls.family] = cls


def split_id(id):
    try:
        a, b, c = (int(x, 16) for x in id.split('.'))
    except ValueError:
        raise NotADevice(id)
    return a, b, c

class _Value:
    def __init__(self, path, typ):
        self.path = path
        self.typ = typ

    def __repr__(self):
        return "<%s: %s %s>" % (self.__class__.__name__,
                                self.path, self.typ)

class _RValue(_Value):
    def __init__(self, path, typ):
        super().__init__(path, typ)
        if typ in {'f', 'g', 'p', 't'}:
            self.conv = float
        elif typ in {'i', 'u'}:
            self.conv = int
        elif typ == 'y':
            self.conv = lambda x: bool(int(x))
        elif typ == 'b':
            self.conv = lambda x: x
        else:
            self.conv = lambda x: x.decode('utf-8')

class _WValue(_Value):
    def __init__(self, path, typ):
        super().__init__(path, typ)
        if typ == 'b':
            self.conv = lambda x: x
        elif typ == 'y':
            self.conv = lambda x: b'1' if x else b'0'
        else:
            self.conv = lambda x: str(x).encode("utf-8")

class SimpleValue(_RValue):
    """Accessor for direct attribute access"""

    async def __get__(slf, self, cls):
        res = await self.dev.attr_get(*slf.path)
        return slf.conv(res)

class SimpleGetter(_RValue):
    """Accessor for get_* function"""

    def __get__(slf, self, cls):
        async def getter():
            res = await self.dev.attr_get(*slf.path)
            return slf.conv(res)

        return getter

class SimpleSetter(_WValue):
    """Accessor for set_* function"""

    def __get__(slf, self, cls):
        async def setter(val):
            await self.dev.attr_set(*slf.path, value=slf.conv(val))

        return setter

class MultiValue(_RValue):
    """Accessor for direct array access"""

    async def __get__(slf, self, cls):
        p = slf.path[:-1] + (slf.path[-1]+'.ALL',)
        res = await self.dev.attr_get(*p)
        conv = slf.conv
        return [conv(v) for v in res.split(b',')]

class MultiGetter(_RValue):
    """Accessor for array get_* function"""

    def __get__(slf, self, cls):
        async def getter():
            p = slf.path[:-1] + (slf.path[-1]+'.ALL',)
            res = await self.dev.attr_get(*p)
            conv = slf.conv
            return [conv(v) for v in res.split(b',')]

        return getter

class MultiSetter(_WValue):
    """Accessor for array set_* function"""

    def __get__(slf, self, cls):
        async def setter(val):
            conv = slf.conv
            p = slf.path[:-1] + (slf.path[-1]+'.ALL',)
            val=b','.join(conv(v) for v in val)
            await self.dev.attr_set(*p, value=val)

        return setter

class ArrayValue(_RValue):
    """Accessor for direct array element access"""
    def __init__(self, path, typ, num):
        super().__init__(path, typ)
        self.num = num

    def __get__(slf, self, cls):
        class IdxObj:
            async def __getitem__(sl, idx):
                if slf.num:
                    idx = str(idx)
                else:
                    idx = chr(ord('A')+idx)
                p = slf.path[:-1] + (slf.path[-1]+'.'+idx,)
                res = await self.dev.attr_get(*p)
                return slf.conv(res)
        return IdxObj()

class ArrayGetter(_RValue):
    """Accessor for array element get_* function"""
    def __init__(self, path, typ, num):
        super().__init__(path, typ)
        self.num = num

    def __get__(slf, self, cls):
        async def getter(idx):
            if slf.num:
                idx = str(idx)
            else:
                idx = chr(ord('A')+idx)
            p = slf.path[:-1] + (slf.path[-1]+'.'+idx,)
            res = await self.dev.attr_get(*p)
            return slf.conv(res)
        return getter

class ArraySetter(_WValue):
    """Accessor for array element set_* function"""
    def __init__(self, path, typ, num):
        super().__init__(path, typ)
        self.num = num

    def __get__(slf, self, cls):
        async def setter(idx, val):
            if slf.num:
                idx = str(idx)
            else:
                idx = chr(ord('A')+idx)
            p = slf.path[:-1] + (slf.path[-1]+'.'+idx,)
            await self.dev.attr_set(*p, value=slf.conv(val))
        return setter

class SubDir:
    _subdirs = set()
    # dev = None  # needs to be filled by subclass
    def __getattr__(self, name):
        if name not in self._subdirs:
            return super().__getattribute__(name)
        c = getattr(self,'_cls_'+name)(self)
        c.dev = self.dev
        return c

async def setup_accessors(server, cls, typ, *subdir):
    cls.fields = {}
    for d in await server.dir("structure", typ, *subdir):
        dd = subdir + (d,)
        try:
            v = await server.attr_get("structure", typ, *dd)
        except IsDirError:

            t = typ
            class SubPath(SubDir):
                typ = t
                subdir = dd
                def __init__(self, base):
                    self.base = base
                def __repr__(self):
                    return "<%s %s %s>" % (self.__class__.__name, self.base, self.subdir)
                def __get__(self, obj, cls):
                    if obj is None:
                        return cls
                    try:
                        return getattr(obj,"_"+self.dd[-1])
                    except AttributeError:
                        c = getattr(cls, '_cls_'+d)()
                        setattr(obj,"_"+self.dd[-1], c)
                        c.dev = obj.dev
                        return c

            SubPath.__name__ = '_cls_'+d
            setattr(cls, '_cls_'+d, SubPath)
            cls._subdirs.add(d)
            await setup_accessors(server, SubPath, typ, *dd)

        else:
            v = v.decode("utf-8").split(",")
            try:
                v[1] = int(v[1])
                v[2] = int(v[2])
                v[4] = int(v[4])
            except ValueError:
                raise ValueError("broken setup vector",(typ,dd),v)
            if v[1] == 0:
                if d.endswith('.0'):
                    num = True
                elif d.endswith('.A'):
                    num = False
                else:
                    num = None
                v[1] = num
                cls.fields[d] = v

                if num is None:
                    if v[3] in {'ro', 'rw'}:
                        if hasattr(cls, d):
                            logger.debug("%s: not overwriting %s" % (cls, d))
                        else:
                            setattr(cls, d, SimpleValue(dd, v[0]))
                        setattr(cls, 'get_' + d, SimpleGetter(dd, v[0]))
                    if v[3] in {'wo', 'rw'}:
                        setattr(cls, 'set_' + d, SimpleSetter(dd, v[0]))
                else:
                    d = d[:-2]
                    dd = subdir + (d,)
                    if v[3] in {'ro', 'rw'}:
                        if hasattr(cls, d):
                            logger.debug("%s: not overwriting %s" % (cls, d))
                        else:
                            setattr(cls, d, ArrayValue(dd, v[0], num))
                        setattr(cls, 'get_' + d, ArrayGetter(dd, v[0], num))
                    if v[3] in {'wo', 'rw'}:
                        setattr(cls, 'set_' + d, ArraySetter(dd, v[0], num))

                    if v[3] in {'ro', 'rw'}:
                        if hasattr(cls, d+'_all'):
                            logger.debug("%s: not overwriting %s" % (cls, d+'_all'))
                        else:
                            setattr(cls, d+'_all', MultiValue(dd, v[0]))
                        setattr(cls, 'get_' + d + '_all', MultiGetter(dd, v[0]))
                    if v[3] in {'wo', 'rw'}:
                        setattr(cls, 'set_' + d + '_all', MultiSetter(dd, v[0]))



class Device(SubDir):
    """Base class for devices.

    A device may or may not have a known location.
    """
    _did_setup = False
    
    def __init__(self, service, id):
        logger.debug("NewDev %s", id)

    @property
    def dev(self):
        return self

    def __new__(cls, service, id):
        family_id, code, chksum = split_id(id)

        cls = dev_classes.get(family_id)
        if cls is None:

            class cls(Device):
                family = family_id

            cls.__name__ = "Device_%02x" % (family_id,)
            dev_classes[family_id] = cls

        self = object.__new__(cls)

        self.id = id.upper()
        self.family = family_id
        self.code = code
        self.chksum = chksum

        self.service = service
        self.bus = None

        self._unseen = 0
        self._events = []
        self._wait_bus = anyio.create_event()

        return self

    def queue_event(self, evt):
        """Remember this event. Used if an event arrives when the device
        hasn't yet been set up by high-level code
        """
        self._events.append(evt)

    @property
    def queued_events(self):
        """Return queued events. Shall be called exactly once."""
        e,self._events = self._events,None
        if e is None:
            raise RuntimeError("You cannot call `queued_events` "
                               "more than once")
        return iter(e)

    @classmethod
    async def setup_struct(cls, server):
        """Read the device's structural data from OWFS
        and add methods to access the fields"""

        if cls._did_setup is not False:
            return
        cls._did_setup = None

        try:
            fc = "%02X" % (cls.family)
            await setup_accessors(server, cls, fc)

        except BaseException:
            cls._did_setup = False
            raise
        else:
            cls._did_setup = True

    def __eq__(self, x):
        x = getattr(x, 'id', x)
        return self.id == x

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return "<%s:%s @ %s>" % (self.__class__.__name__, self.id, self.bus)

    def buses(self):
        return set()

    async def locate(self, bus):
        """The device has been seen here."""
        if self.bus is bus:
            return
        self.bus = bus
        await self._wait_bus.set()
        await self.service.push_event(DeviceLocated(self))

    async def wait_bus(self):
        await self._wait_bus.wait()

    async def delocate(self, bus):
        """The device is no longer located here."""
        if self.bus is bus:
            self._wait_bus = anyio.create_event()
            await self._delocate()

    async def _delocate(self):
        await self.bus._del_device(self)
        self.bus = None
        await self.service.push_event(DeviceNotFound(self))

    async def attr_get(self, *attr: List[str]):
        """Read this attribute"""
        if self.bus is None:
            raise NoLocationKnown(self)
        return await self.bus.attr_get(self.id, *attr)

    async def attr_set(self, *attr: List[str], value):
        """Write this attribute"""
        if self.bus is None:
            raise NoLocationKnown(self)
        return await self.bus.attr_set(self.id, *attr, value=value)

    def polling_items(self):
        """Enumerate poll variants supported by this device.
        
        This is a generator. If you override, call::

            yield from super().polling_items()

        See the associated ``poll_<name>`` methods on
        :class:`anyio_owfs.bus.Bus` for details.

        Special return values:

        * "alarm": you need to implement ``.stop_alarm``
        """
        if False:
            yield None

    def polling_interval(self, typ: str):
        """Return the interval WRT how often to poll for this type.

        The default implementation looks up the "interval_<typ>" attribute
        or returns ``None`` if that doesn't exist.
        """
        return getattr(self, "interval_"+typ, None)
    
    async def set_polling_interval(self, typ: str, value: Optional[float]):
        setattr(self, "interval_"+typ, value)
        if self.bus is not None:
            await self.bus.update_poll()

    async def poll_alarm(self):
        """Tells the device not to trigger an alarm any more.

        You *need* to override this if your device can trigger an alarm
        condition. Also, this method *must* disable the alarm; your
        application can re-enable it later, when processing the
        :class:`anyio_owfs.event.DeviceAlarm` event.
        """
        raise NotImplementedError("<%s> needs 'stop_alarm'" % (self.__class__.__name__,))

@register
class SwitchDevice(Device):
    family = 0x1F

    def buses(self):
        """Return a list of the connected buses.
        'main' should be processed first.
        """
        b = []
        b.append((self.id, "main"))
        b.append((self.id, "aux"))
        return b

    async def poll_alarm(self):
        """Clear alarm"""
        res = await self.get_event_all()
        await self.set_clearalarm(1)

@register
class TemperatureDevice(Device):
    family = 0x10

    interval_temperature = None

    async def poll_alarm(self):
        """Turn off alarm condition by adapting the temperature bounds
        """
        self.alarm_temperature = t = await self.latesttemp
        reasons = {'temp':t}

        t_h = await self.temphigh
        if t > t_h:
            await self.set_temphigh(int(t + 2))
            reasons['high'] = t_h
        t_l = await self.templow
        if t < t_l:
            await self.set_templow(int(t - 1))
            reasons['low'] = t_l
        return reasons

    def polling_items(self):
        yield from super().polling_items()
        yield "temperature"
        yield "alarm"

    async def poll_temperature(self):
        t = await self.latesttemp
        await self.service.push_event(DeviceValue(self, "temperature", t))

    @property
    def temperature(self):
        return self.latesttemp

@register
class TemperatureBDevice(Device):
    family = 0x28

@register
class VoltageDevice(Device):
    family = 0x20

    interval_voltage = None
    alarm_voltage = None

    async def poll_alarm(self):
        """Turn off alarm condition by adapting the voltage bounds
        """
        reasons = {}
        v = await self.voltage_all
        ah = await self.alarm.high_all
        al = await self.alarm.low_all
        pow = await self.set_alarm.unset
        if pow:
            reasons['power_on'] = True
            pow = await self.set_alarm.set_unset(0)
        if any(ah):
            vh = await self.set_alarm.set_high_all
            for i in range(4):
                if not ah[i]:
                    continue
                reasons['high_'+str(i)] = vh[i]
                await self.set_alarm.set_high(i,0)
        if any(al):
            vl = await self.set_alarm.set_low_all
            for i in range(4):
                if not al[i]:
                    continue
                reasons['low_'+str(i)] = vl[i]
                await self.set_alarm.set_low(i,0)
        for i in range(4):
            reasons['volt_'+str(i)] = v[i]

        return reasons

    def polling_items(self):
        yield from super().polling_items()
        yield "voltage"
        yield "alarm"

    async def poll_voltage(self):
        v = await self.volt_all
        await self.service.push_event(DeviceValue(self, "volt_all", v))




# for message type classification see
# http://owfs.org/index.php?page=owserver-message-types
# and 'enum msg_classification' from module/owlib/src/include/ow_message.h

import attr

import logging
logger = logging.getLogger(__name__)


class OWFSReplyError(Exception):
    pass


@attr.s(cmp=False)
class GenericOWFSReplyError(OWFSReplyError):
    err = attr.ib()
    req = attr.ib()
    server = attr.ib()


_errors = {}


@attr.s(cmp=False)
class OWFSReplyError_(OWFSReplyError):
    req = attr.ib()
    server = attr.ib()


def _register(cls):
    _errors[cls.err] = cls
    return cls


@_register
@attr.s(cmp=False)
class PermissionError(OWFSReplyError_):
    err = 13


@_register
@attr.s(cmp=False)
class InUseError(OWFSReplyError_):
    err = 98


@_register
@attr.s(cmp=False)
class NotAvailableError(OWFSReplyError_):
    err = 99


@_register
@attr.s(cmp=False)
class TryAgainError(OWFSReplyError_):
    err = 11


@_register
@attr.s(cmp=False)
class BadFSError(OWFSReplyError_):
    err = 9


@_register
@attr.s(cmp=False)
class BadMsgError(OWFSReplyError_):
    err = 74


@_register
@attr.s(cmp=False)
class BusyError(OWFSReplyError_):
    err = 16


@_register
@attr.s(cmp=False)
class ConnAbortedError(OWFSReplyError_):
    err = 103


@_register
@attr.s(cmp=False)
class FaultError(OWFSReplyError_):
    err = 14


@_register
@attr.s(cmp=False)
class InterruptedError(OWFSReplyError_):
    err = 4


@_register
@attr.s(cmp=False)
class InvalidDataError(OWFSReplyError_):
    err = 22


@_register
@attr.s(cmp=False)
class BusIOError(OWFSReplyError_):
    err = 5


@_register
@attr.s(cmp=False)
class IsDirError(OWFSReplyError_):
    err = 21


@_register
@attr.s(cmp=False)
class LoopError(OWFSReplyError_):
    err = 40


@_register
@attr.s(cmp=False)
class MsgSizeError(OWFSReplyError_):
    err = 90


@_register
@attr.s(cmp=False)
class NameTooLongError(OWFSReplyError_):
    err = 36


@_register
@attr.s(cmp=False)
class NoBufsError(OWFSReplyError_):
    err = 105


@_register
@attr.s(cmp=False)
class NoDeviceError(OWFSReplyError_):
    err = 19


@_register
@attr.s(cmp=False)
class NoEntryError(OWFSReplyError_):
    err = 2


@_register
@attr.s(cmp=False)
class NoFreeMemoryError(OWFSReplyError_):
    err = 12


@_register
@attr.s(cmp=False)
class NoMessageError(OWFSReplyError_):
    err = 42


@_register
@attr.s(cmp=False)
class NoDirectoryError(OWFSReplyError_):
    err = 20


@_register
@attr.s(cmp=False)
class NotSupportedError(OWFSReplyError_):
    err = 95


@_register
@attr.s(cmp=False)
class RangeError(OWFSReplyError_):
    err = 34


@_register
@attr.s(cmp=False)
class ReadOnlyError(OWFSReplyError_):
    err = 30


@_register
@attr.s(cmp=False)
class InputPathTooLongError(OWFSReplyError_):
    err = 26


@_register
@attr.s(cmp=False)
class BadPathSyntaxError(OWFSReplyError_):
    err = 27


@_register
@attr.s(cmp=False)
class TextInPathError(OWFSReplyError_):
    err = 77


@_register
@attr.s(cmp=False)
class BadCRC8Error(OWFSReplyError_):
    err = 28


@_register
@attr.s(cmp=False)
class UnknownNameError(OWFSReplyError_):
    err = 29


@_register
@attr.s(cmp=False)
class AliasTooLongError(OWFSReplyError_):
    err = 31


@_register
@attr.s(cmp=False)
class UnknownPropertyError(OWFSReplyError_):
    err = 32


@_register
@attr.s(cmp=False)
class NotAnArrayError(OWFSReplyError_):
    err = 33


@_register
@attr.s(cmp=False)
class IsAnArrayError(OWFSReplyError_):
    err = 35


@_register
@attr.s(cmp=False)
class NotBitfieldError(OWFSReplyError_):
    err = 37


@_register
@attr.s(cmp=False)
class IndexTooLargeError(OWFSReplyError_):
    err = 38


@_register
@attr.s(cmp=False)
class NoSubpathError(OWFSReplyError_):
    err = 39


@_register
@attr.s(cmp=False)
class DeviceNotFoundError(OWFSReplyError_):
    err = 41


@_register
@attr.s(cmp=False)
class DeviceError(OWFSReplyError_):
    err = 43


@_register
@attr.s(cmp=False)
class BusShortError(OWFSReplyError_):
    err = 44


@_register
@attr.s(cmp=False)
class NoSuchBusError(OWFSReplyError_):
    err = 45


@_register
@attr.s(cmp=False)
class BusNotAppropriateError(OWFSReplyError_):
    err = 46


@_register
@attr.s(cmp=False)
class BusNotRespondingError(OWFSReplyError_):
    err = 47


@_register
@attr.s(cmp=False)
class BusResetError(OWFSReplyError_):
    err = 48


@_register
@attr.s(cmp=False)
class BusClosedError(OWFSReplyError_):
    err = 49


@_register
@attr.s(cmp=False)
class BusNotOpenedError(OWFSReplyError_):
    err = 50


@_register
@attr.s(cmp=False)
class BusCommunicationError(OWFSReplyError_):
    err = 51


@_register
@attr.s(cmp=False)
class BusTimeoutError(OWFSReplyError_):
    err = 52


@_register
@attr.s(cmp=False)
class TelnetError(OWFSReplyError_):
    err = 53


@_register
@attr.s(cmp=False)
class TCPError(OWFSReplyError_):
    err = 54


@_register
@attr.s(cmp=False)
class BusIsLocalError(OWFSReplyError_):
    err = 55


@_register
@attr.s(cmp=False)
class BusIsRemoteError(OWFSReplyError_):
    err = 56


@_register
@attr.s(cmp=False)
class ReadTooLargeError(OWFSReplyError_):
    err = 57


@_register
@attr.s(cmp=False)
class DataCommunicationError(OWFSReplyError_):
    err = 58


@_register
@attr.s(cmp=False)
class NotRPropertyError(OWFSReplyError_):
    err = 59


@_register
@attr.s(cmp=False)
class NotReadablePropertyError(OWFSReplyError_):
    err = 60


@_register
@attr.s(cmp=False)
class DataTooLargeError(OWFSReplyError_):
    err = 61


@_register
@attr.s(cmp=False)
class DataTooSmallError(OWFSReplyError_):
    err = 62


@_register
@attr.s(cmp=False)
class DataFormatError(OWFSReplyError_):
    err = 63


@_register
@attr.s(cmp=False)
class NotWPropertyError(OWFSReplyError_):
    err = 64


@_register
@attr.s(cmp=False)
class NotWritablePropertyError(OWFSReplyError_):
    err = 65


@_register
@attr.s(cmp=False)
class ReadOnlyModeError(OWFSReplyError_):
    err = 66


@_register
@attr.s(cmp=False)
class DataCommError(OWFSReplyError_):
    err = 67


@_register
@attr.s(cmp=False)
class OutputPathTooLongError(OWFSReplyError_):
    err = 68


@_register
@attr.s(cmp=False)
class NotADirectoryError(OWFSReplyError_):
    err = 69


@_register
@attr.s(cmp=False)
class NotADeviceError(OWFSReplyError_):
    err = 70


@_register
@attr.s(cmp=False)
class UnknownQueryError(OWFSReplyError_):
    err = 71


@_register
@attr.s(cmp=False)
class SocketError(OWFSReplyError_):
    err = 72


@_register
@attr.s(cmp=False)
class TimeoutError(OWFSReplyError_):
    err = 73


@_register
@attr.s(cmp=False)
class VersionError(OWFSReplyError_):
    err = 75


@_register
@attr.s(cmp=False)
class PacketSizeError(OWFSReplyError_):
    err = 76


@_register
@attr.s(cmp=False)
class UnexpectedNullError(OWFSReplyError_):
    err = 78


@_register
@attr.s(cmp=False)
class NoMemoryError(OWFSReplyError_):
    err = 79



import trio
from asyncowfs import OWFS
from asyncowfs.protocol import MessageProtocol, OWMsg
from asyncowfs.error import OWFSReplyError, NoEntryError, IsDirError
try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager
from functools import partial

import logging
logger = logging.getLogger(__name__)


def _chk(v):
    if v is None:
        return False
    v[0] += 1
    if v[0] >= len(v):
        v[0] = 1
    return v[v[0]]


async def _schk(v):
    if v is None:
        return
    v[0] += 1
    if v[0] >= len(v):
        v[0] = 1
    if v[v[0]] > 0:
        logger.debug("Slow reply %s", v[v[0]])
    await trio.sleep(v[v[0]])


class FakeMaster:
    def __init__(self, stream):
        self.stream = stream

async def some_server(tree, msgs, options, socket):
    rdr = MessageProtocol(FakeMaster(socket), is_server=True)
    logger.debug("START Server")
    if msgs:
        midx = 0
        if isinstance(msgs[0], list):
            msgs.insert(0, 0)
        if isinstance(msgs[0], int):
            midx = msgs[0]
            msgs[0] += 1
            msgs = msgs[msgs[0]]
        msgs = iter(msgs)
        mpos = 0

    def _end(msgs):
        nonlocal mpos
        if msgs:
            mpos += 1
            try:
                m = next(msgs)
            except StopIteration:
                pass
            else:
                raise RuntimeError("Message '%s' not seen (%d,%d)" % (m, midx, mpos))

    each_busy = options.get("busy_every", None)
    each_close = options.get("close_every", None)
    each_slow = options.get("slow_every", None)
    try:
        if _chk(each_close):
            _end(msgs)
            return
        async for command, format_flags, data, offset in rdr:
            try:
                await _schk(each_slow)
                if _chk(each_busy):
                    await rdr.write(0, format_flags, 0, data=None)
                    continue
                if msgs:
                    try:
                        m = next(msgs)
                    except StopIteration:
                        raise RuntimeError(
                            "Unexpected command %d/%d: %d %x %s %d" %
                            (midx, mpos, command, format_flags, repr(data), offset)
                        ) from None
                    mpos += 1
                    try:
                        m._check(command, data)
                    except Exception:
                        raise RuntimeError("Message '%s' wrong seen (%d,%d)" % (m, midx, mpos))
                if _chk(each_close):
                    _end(msgs)
                    return
                if command == OWMsg.nop:
                    await rdr.write(0, format_flags, 0)
                elif command == OWMsg.dirall:
                    data = data.rstrip(b'\0')
                    subtree = tree
                    path = []
                    for k in data.split(b'/'):
                        if k == b'':
                            continue
                        path.append(k)
                        k = k.decode("utf-8")
                        try:
                            subtree = subtree[k]
                        except KeyError:
                            raise NoEntryError(command, data)
                    res = []
                    for k, v in sorted(subtree.items()):
                        k = k.encode('utf-8')
                        res.append(k)
                    if path:
                        path = b'/' + b'/'.join(path) + b'/'
                    else:
                        path = b'/'
                    data = b','.join(path + k for k in res)
                    await rdr.write(0, format_flags, len(data), data + b'\0')
                elif command == OWMsg.read:
                    data = data.rstrip(b'\0')
                    res = tree
                    for k in data.split(b'/'):
                        if k == b'':
                            continue
                        k = k.decode("utf-8")
                        try:
                            res = res[k]
                        except KeyError:
                            raise NoEntryError(command, data)
                    if isinstance(res, dict):
                        raise IsDirError(command, data)
                    if not isinstance(res, bytes):
                        res = str(res).encode('utf-8')
                    await rdr.write(0, format_flags, len(res), res + b'\0')
                elif command == OWMsg.write:
                    val = data[-offset:].decode("utf-8")
                    data = data[:-offset]
                    data = data.rstrip(b'\0')
                    res = tree
                    last = None
                    for k in data.split(b'/'):
                        if k == b'':
                            continue
                        if last is not None:
                            try:
                                res = res[last]
                            except KeyError:
                                raise NoEntryError(command, data)
                        last = k.decode("utf-8")
                    assert last is not None
                    if last not in res:
                        raise NoEntryError(command, data)
                    res[last] = val
                    await rdr.write(0, format_flags, 0)
                else:
                    raise RuntimeError(
                        "Unknown command: %d %x %s %d" %
                        (command, format_flags, repr(data), offset)
                    )
            except OWFSReplyError as err:
                logger.info("Error: %s", err)
                await rdr.write(-err.err, format_flags)

        _end(msgs)
    except trio.BrokenResourceError:
        _end(msgs)
    finally:
        logger.debug("END Server")


class EventChecker:
    def __init__(self, events=[]):
        self.events = events[:]

    def add(self, event):
        self.events.push(event)

    async def __call__(self, ow, task_status=trio.TASK_STATUS_IGNORED):
        self.pos = 0
        try:
            async with ow.events as ev:
                task_status.started()
                async for e in ev:
                    logger.debug("Event %s", e)
                    self.check_next(e)
        except RuntimeError:
            raise
        except Exception as exc:
            self.check_last()
        else:
            self.check_last()
        # don't check on BaseException =Cancellation

    def check_next(self, e):
        try:
            t = self.events[self.pos]
        except IndexError:
            raise RuntimeError("Unexpected event %s" % (e,))
        self.pos += 1
        if isinstance(t, type) and isinstance(e, t):
            pass
        elif not (t == e):
            raise RuntimeError("Wrong event: want %s but has %s" % (t, e))

    def check_last(self):
        logger.debug("Event END")
        if self.pos != len(self.events):
            raise RuntimeError("Superfluous event #%d: %s" % (self.pos, self.events[self.pos]))


@asynccontextmanager
async def server(tree={}, msgs=(), options={}, events=None, polling=False, scan=None, initial_scan=True, **kw):
    async with OWFS(**kw) as ow:
        async with trio.open_nursery() as n:
            s = None
            try:
                server = await n.start(
                    partial(trio.serve_tcp, host="127.0.0.1"),
                    partial(some_server, tree, msgs, options), 0
                )
                if events is not None:
                    await n.start(events, ow)
                addr = server[0].socket.getsockname()

                s = await ow.add_server(*addr, polling=polling, scan=scan, initial_scan=initial_scan)
                ow.test_server = s
                yield ow
            finally:
                ow.test_server = None
                with trio.CancelScope(shield=True):
                    if s is not None:
                        await s.drop()
                    await ow.push_event(None)
                    await trio.sleep(0.1)
                n.cancel_scope.cancel()

